from typing import Optional

from freeplay.errors import FreeplayConfigurationError
from freeplay.resources.customer_feedback import CustomerFeedback
from freeplay.resources.prompts import Prompts, APITemplateResolver, TemplateResolver
from freeplay.resources.recordings import Recordings
from freeplay.resources.sessions import Sessions
from freeplay.resources.test_runs import TestRuns
from freeplay.support import CallSupport


class Freeplay:
    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str,
            template_resolver: Optional[TemplateResolver] = None
    ) -> None:
        if not freeplay_api_key or not freeplay_api_key.strip():
            raise FreeplayConfigurationError("Freeplay API key not set. It must be set to the Freeplay API.")

        self.call_support = CallSupport(
            freeplay_api_key,
            api_base
        )
        self.freeplay_api_key = freeplay_api_key
        self.api_base = api_base

        resolver: TemplateResolver
        if template_resolver is None:
            resolver = APITemplateResolver(self.call_support)
        else:
            resolver = template_resolver

        # Resources ========
        self.customer_feedback = CustomerFeedback(self.call_support)
        self.prompts = Prompts(self.call_support, resolver)
        self.recordings = Recordings(self.call_support)
        self.sessions = Sessions()
        self.test_runs = TestRuns(self.call_support)



from .freeplay import Freeplay
from .resources.prompts import PromptInfo
from .resources.recordings import CallInfo, ResponseInfo, RecordPayload, TestRunInfo
from .resources.sessions import SessionInfo

__all__ = [
    'CallInfo',
    'Freeplay',
    'PromptInfo',
    'RecordPayload',
    'ResponseInfo',
    'SessionInfo',
    'TestRunInfo',
]



from requests import Response


class FreeplayError(Exception):
    pass


class FreeplayConfigurationError(FreeplayError):
    pass


class FreeplayClientError(FreeplayError):
    pass


class FreeplayServerError(FreeplayError):
    pass


class LLMClientError(FreeplayError):
    pass


class LLMServerError(FreeplayError):
    pass


def freeplay_response_error(message: str, response: Response) -> FreeplayError:
    full_message = f'{message} [{response.status_code}]'

    if response.status_code in range(400, 500):
        return FreeplayClientError(full_message)
    else:
        return FreeplayServerError(full_message)



import uuid
from dataclasses import dataclass
from typing import Optional, Dict, Union

CustomMetadata = Optional[Dict[str, Union[str, int, float]]]


@dataclass
class SessionInfo:
    session_id: str
    custom_metadata: CustomMetadata


@dataclass
class Session:
    session_id: str
    custom_metadata: CustomMetadata

    def __init__(self, session_id: str, custom_metadata: CustomMetadata):
        self.session_id = session_id
        self.custom_metadata = custom_metadata
        self._session_info = SessionInfo(self.session_id, self.custom_metadata)

    @property
    def session_info(self) -> SessionInfo:
        return self._session_info


class Sessions:
    # noinspection PyMethodMayBeStatic
    def create(self, custom_metadata: CustomMetadata = None) -> Session:
        return Session(session_id=str(uuid.uuid4()), custom_metadata=custom_metadata)



import json
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from requests import HTTPError

from freeplay import api_support
from freeplay.errors import FreeplayClientError, FreeplayError
from freeplay.llm_parameters import LLMParameters
from freeplay.model import InputVariables, OpenAIFunctionCall
from freeplay.resources.prompts import PromptInfo
from freeplay.resources.sessions import SessionInfo
from freeplay.support import CallSupport

logger = logging.getLogger(__name__)


@dataclass
class CallInfo:
    provider: str
    model: str
    start_time: float
    end_time: float
    model_parameters: LLMParameters
    provider_info: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_prompt_info(prompt_info: PromptInfo, start_time: float, end_time: float) -> 'CallInfo':
        return CallInfo(
            provider=prompt_info.provider,
            model=prompt_info.model,
            start_time=start_time,
            end_time=end_time,
            model_parameters=prompt_info.model_parameters,
            provider_info=prompt_info.provider_info,
        )


@dataclass
class ResponseInfo:
    is_complete: bool
    function_call_response: Optional[OpenAIFunctionCall] = None
    prompt_tokens: Optional[int] = None
    response_tokens: Optional[int] = None


@dataclass
class TestRunInfo:
    test_run_id: str
    test_case_id: str


@dataclass
class RecordPayload:
    all_messages: List[Dict[str, str]]
    inputs: InputVariables

    session_info: SessionInfo
    prompt_info: PromptInfo
    call_info: CallInfo
    response_info: ResponseInfo
    test_run_info: Optional[TestRunInfo] = None


@dataclass
class RecordResponse:
    completion_id: str


class Recordings:
    def __init__(self, call_support: CallSupport):
        self.call_support = call_support

    def create(self, record_payload: RecordPayload) -> RecordResponse:
        if len(record_payload.all_messages) < 1:
            raise FreeplayClientError("Messages list must have at least one message. "
                                      "The last message should be the current response.")

        completion = record_payload.all_messages[-1]
        history_as_string = json.dumps(record_payload.all_messages[0:-1])

        record_api_payload = {
            "session_id": record_payload.session_info.session_id,
            "prompt_template_id": record_payload.prompt_info.prompt_template_id,
            "project_version_id": record_payload.prompt_info.prompt_template_version_id,
            "start_time": record_payload.call_info.start_time,
            "end_time": record_payload.call_info.end_time,
            "tag": record_payload.prompt_info.environment,
            "inputs": record_payload.inputs,
            "prompt_content": history_as_string,
            # Content may not be set for function calls, but it is required in the record API payload.
            "return_content": completion.get('content', ''),
            "format_type": None,
            "is_complete": record_payload.response_info.is_complete,
            "model": record_payload.call_info.model,
            "provider": record_payload.call_info.provider,
            "llm_parameters": record_payload.call_info.model_parameters,
            "provider_info": record_payload.call_info.provider_info,
        }

        if record_payload.session_info.custom_metadata is not None:
            record_api_payload['custom_metadata'] = record_payload.session_info.custom_metadata

        if record_payload.response_info.function_call_response is not None:
            record_api_payload['function_call_response'] = record_payload.response_info.function_call_response

        if record_payload.test_run_info is not None:
            record_api_payload['test_run_id'] = record_payload.test_run_info.test_run_id

        if record_payload.test_run_info is not None:
            record_api_payload['test_case_id'] = record_payload.test_run_info.test_case_id

        try:
            recorded_response = api_support.post_raw(
                api_key=self.call_support.freeplay_api_key,
                url=f'{self.call_support.api_base}/v1/record',
                payload=record_api_payload
            )
            recorded_response.raise_for_status()
            json_dom = recorded_response.json()
            return RecordResponse(completion_id=str(json_dom['completion_id']))
        except HTTPError as e:
            message = f'There was an error recording to Freeplay. Call will not be logged. ' \
                      f'Status: {e.response.status_code}. '

            if e.response.content:
                try:
                    content = e.response.content
                    json_body = json.loads(content)
                    if 'message' in json_body:
                        message += json_body['message']
                except:
                    pass
            else:
                message += f'{e.__class__}'

            raise FreeplayError(message) from e

        except Exception as e:
            status_code = -1
            if hasattr(e, 'response') and hasattr(e.response, 'status_code'):
                status_code = e.response.status_code

            message = f'There was an error recording to Freeplay. Call will not be logged. ' \
                      f'Status: {status_code}. {e.__class__}'

            raise FreeplayError(message) from e



from dataclasses import dataclass
from typing import List, Optional

from freeplay.model import InputVariables
from freeplay.resources.recordings import TestRunInfo
from freeplay.support import CallSupport


@dataclass
class TestCase:
    def __init__(
            self,
            test_case_id: str,
            variables: InputVariables,
            output: Optional[str],
    ):
        self.id = test_case_id
        self.variables = variables
        self.output = output


@dataclass
class TestRun:
    def __init__(
            self,
            test_run_id: str,
            test_cases: List[TestCase]
    ):
        self.test_run_id = test_run_id
        self.test_cases = test_cases

    def get_test_cases(self) -> List[TestCase]:
        return self.test_cases

    def get_test_run_info(self, test_case_id: str) -> TestRunInfo:
        return TestRunInfo(self.test_run_id, test_case_id)


class TestRuns:
    def __init__(self, call_support: CallSupport) -> None:
        self.call_support = call_support

    def create(self, project_id: str, testlist: str, include_outputs: bool = False) -> TestRun:
        test_run = self.call_support.create_test_run(project_id, testlist, include_outputs)
        test_cases = [
            TestCase(test_case_id=test_case.id, variables=test_case.variables, output=test_case.output)
            for test_case in test_run.test_cases
        ]

        return TestRun(test_run.test_run_id, test_cases)



from dataclasses import dataclass
from typing import List, Union, Any, Dict, Mapping, TypedDict

InputValue = Union[str, int, bool, Dict[str, Any], List[Any]]
InputVariables = Mapping[str, InputValue]
TestRunInput = Mapping[str, InputValue]


@dataclass
class TestRun:
    id: str
    inputs: List[TestRunInput]


class OpenAIFunctionCall(TypedDict):
    name: str
    arguments: str



from dataclasses import dataclass
from typing import Dict, Union

from freeplay.support import CallSupport


@dataclass
class CustomerFeedbackResponse:
    pass


class CustomerFeedback:
    def __init__(self, call_support: CallSupport) -> None:
        self.call_support = call_support

    def update(self, completion_id: str, feedback: Dict[str, Union[bool, str, int, float]]) -> CustomerFeedbackResponse:
        self.call_support.update_customer_feedback(completion_id, feedback)
        return CustomerFeedbackResponse()



from dataclasses import dataclass
from json import JSONEncoder
from typing import Optional, Dict, Any, List, Union

from freeplay import api_support
from freeplay.api_support import try_decode
from freeplay.errors import freeplay_response_error, FreeplayServerError
from freeplay.model import InputVariables


@dataclass
class PromptTemplateMetadata:
    provider: Optional[str]
    flavor: Optional[str]
    model: Optional[str]
    params: Optional[Dict[str, Any]] = None
    provider_info: Optional[Dict[str, Any]] = None


@dataclass
class PromptTemplate:
    prompt_template_id: str
    prompt_template_version_id: str
    prompt_template_name: str
    content: List[Dict[str, str]]
    metadata: PromptTemplateMetadata
    format_version: int


@dataclass
class PromptTemplates:
    prompt_templates: List[PromptTemplate]


class PromptTemplateEncoder(JSONEncoder):
    def default(self, prompt_template: PromptTemplate) -> Dict[str, Any]:
        return prompt_template.__dict__


class TestCaseTestRunResponse:
    def __init__(self, test_case: Dict[str, Any]):
        self.variables: InputVariables = test_case['variables']
        self.id: str = test_case['id']
        self.output: Optional[str] = test_case.get('output')


class TestRunResponse:
    def __init__(
            self,
            test_run_id: str,
            test_cases: List[Dict[str, Any]]
    ):
        self.test_cases = [
            TestCaseTestRunResponse(test_case)
            for test_case in test_cases
        ]
        self.test_run_id = test_run_id


class CallSupport:
    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str
    ) -> None:
        self.api_base = api_base
        self.freeplay_api_key = freeplay_api_key

    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        response = api_support.get_raw(
            api_key=self.freeplay_api_key,
            url=f'{self.api_base}/v2/projects/{project_id}/prompt-templates/all/{environment}'
        )

        if response.status_code != 200:
            raise freeplay_response_error("Error getting prompt templates", response)

        maybe_prompts = try_decode(PromptTemplates, response.content)
        if maybe_prompts is None:
            raise FreeplayServerError('Failed to parse prompt templates from server')

        return maybe_prompts

    def get_prompt(self, project_id: str, template_name: str, environment: str) -> PromptTemplate:
        response = api_support.get_raw(
            api_key=self.freeplay_api_key,
            url=f'{self.api_base}/v2/projects/{project_id}/prompt-templates/name/{template_name}',
            params={
                'environment': environment
            }
        )

        if response.status_code != 200:
            raise freeplay_response_error(
                f"Error getting prompt template {template_name} in project {project_id} "
                f"and environment {environment}",
                response
            )

        maybe_prompt = try_decode(PromptTemplate, response.content)
        if maybe_prompt is None:
            raise FreeplayServerError(
                f"Error handling prompt {template_name} in project {project_id} "
                f"and environment {environment}"
            )

        return maybe_prompt

    def update_customer_feedback(
            self,
            completion_id: str,
            feedback: Dict[str, Union[bool, str, int, float]]
    ) -> None:
        response = api_support.put_raw(
            self.freeplay_api_key,
            f'{self.api_base}/v1/completion_feedback/{completion_id}',
            feedback
        )
        if response.status_code != 201:
            raise freeplay_response_error("Error updating customer feedback", response)

    def create_test_run(
            self,
            project_id: str,
            testlist: str,
            include_test_case_outputs: bool = False
    ) -> TestRunResponse:
        response = api_support.post_raw(
            api_key=self.freeplay_api_key,
            url=f'{self.api_base}/projects/{project_id}/test-runs-cases',
            payload={
                'testlist_name': testlist,
                'include_test_case_outputs': include_test_case_outputs,
            },
        )

        if response.status_code != 201:
            raise freeplay_response_error('Error while creating a test run.', response)

        json_dom = response.json()

        return TestRunResponse(json_dom['test_run_id'], json_dom['test_cases'])


import copy
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, List, cast, Any, Union

from freeplay.errors import FreeplayConfigurationError, FreeplayClientError
from freeplay.llm_parameters import LLMParameters
from freeplay.model import InputVariables
from freeplay.support import CallSupport
from freeplay.support import PromptTemplate, PromptTemplates, PromptTemplateMetadata
from freeplay.utils import bind_template_variables


class MissingFlavorError(FreeplayConfigurationError):
    def __init__(self, flavor_name: str):
        super().__init__(
            f'Configured flavor ({flavor_name}) not found in SDK. Please update your SDK version or configure '
            'a different model in the Freeplay UI.'
        )


# SDK-Exposed Classes
@dataclass
class PromptInfo:
    prompt_template_id: str
    prompt_template_version_id: str
    template_name: str
    environment: str
    model_parameters: LLMParameters
    provider_info: Optional[Dict[str, Any]]
    provider: str
    model: str
    flavor_name: str


class FormattedPrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[Dict[str, str]],
            formatted_prompt: Optional[List[Dict[str, str]]] = None,
            formatted_prompt_text: Optional[str] = None
    ):
        self.prompt_info = prompt_info
        self.llm_prompt = formatted_prompt
        if formatted_prompt_text:
            self.llm_prompt_text = formatted_prompt_text

        maybe_system_content = next(
            (message['content'] for message in messages if message['role'] == 'system'), None)
        self.system_content = maybe_system_content

        # Note: messages are **not formatted** for the provider.
        self.messages = messages

    def all_messages(
            self,
            new_message: Dict[str, str]
    ) -> List[Dict[str, str]]:
        return self.messages + [new_message]


class BoundPrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[Dict[str, str]]
    ):
        self.prompt_info = prompt_info
        self.messages = messages

    @staticmethod
    def __format_messages_for_flavor(
            flavor_name: str,
            messages: List[Dict[str, str]]
    ) -> Union[str, List[Dict[str, str]]]:
        if flavor_name == 'azure_openai_chat' or flavor_name == 'openai_chat':
            # We need a deepcopy here to avoid referential equality with the llm_prompt
            return copy.deepcopy(messages)
        elif flavor_name == 'anthropic_chat':
            messages_without_system = [message for message in messages if message['role'] != 'system']
            return messages_without_system
        elif flavor_name == 'llama_3_chat':
            if len(messages) < 1:
                raise ValueError("Must have at least one message to format")

            formatted = "<|begin_of_text|>"
            for message in messages:
                formatted += f"<|start_header_id|>{message['role']}<|end_header_id|>\n{message['content']}<|eot_id|>"
            formatted += "<|start_header_id|>assistant<|end_header_id|>"

            return formatted

        raise MissingFlavorError(flavor_name)

    def format(
            self,
            flavor_name: Optional[str] = None
    ) -> FormattedPrompt:
        final_flavor = flavor_name or self.prompt_info.flavor_name
        formatted_prompt = BoundPrompt.__format_messages_for_flavor(final_flavor, self.messages)

        if isinstance(formatted_prompt, str):
            return FormattedPrompt(
                prompt_info=self.prompt_info,
                messages=self.messages,
                formatted_prompt_text=formatted_prompt
            )
        else:
            return FormattedPrompt(
                prompt_info=self.prompt_info,
                messages=self.messages,
                formatted_prompt=formatted_prompt
            )


class TemplatePrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[Dict[str, str]]
    ):
        self.prompt_info = prompt_info
        self.messages = messages

    def bind(self, variables: InputVariables) -> BoundPrompt:
        bound_messages = [
            {'role': message['role'], 'content': bind_template_variables(message['content'], variables)}
            for message in self.messages
        ]
        return BoundPrompt(self.prompt_info, bound_messages)


class TemplateResolver(ABC):
    @abstractmethod
    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        pass

    @abstractmethod
    def get_prompt(self, project_id: str, template_name: str, environment: str) -> PromptTemplate:
        pass


class FilesystemTemplateResolver(TemplateResolver):
    # If you think you need a change here, be sure to check the server as the translations must match. Once we have
    # all the SDKs and all customers on the new common format, this translation can go away.
    __role_translations = {
        'system': 'system',
        'user': 'user',
        'assistant': 'assistant',
        'Assistant': 'assistant',
        'Human': 'user'  # Don't think we ever store this, but in case...
    }

    def __init__(self, freeplay_directory: Path):
        FilesystemTemplateResolver.__validate_freeplay_directory(freeplay_directory)
        self.prompts_directory = freeplay_directory / "freeplay" / "prompts"

    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        self.__validate_prompt_directory(project_id, environment)

        directory = self.prompts_directory / project_id / environment
        prompt_file_paths = directory.glob("*.json")

        prompt_list = []
        for prompt_file_path in prompt_file_paths:
            json_dom = json.loads(prompt_file_path.read_text())
            prompt_list.append(self.__render_into_v2(json_dom))

        return PromptTemplates(prompt_list)

    def get_prompt(self, project_id: str, template_name: str, environment: str) -> PromptTemplate:
        self.__validate_prompt_directory(project_id, environment)

        expected_file: Path = self.prompts_directory / project_id / environment / f"{template_name}.json"

        if not expected_file.exists():
            raise FreeplayClientError(
                f"Could not find prompt with name {template_name} for project "
                f"{project_id} in environment {environment}"
            )

        json_dom = json.loads(expected_file.read_text())
        return self.__render_into_v2(json_dom)

    @staticmethod
    def __render_into_v2(json_dom: Dict[str, Any]) -> PromptTemplate:
        format_version = json_dom.get('format_version')

        if format_version == 2:
            metadata = json_dom['metadata']
            flavor_name = metadata.get('flavor')
            model = metadata.get('model')

            return PromptTemplate(
                format_version=2,
                prompt_template_id=json_dom.get('prompt_template_id'),  # type: ignore
                prompt_template_version_id=json_dom.get('prompt_template_version_id'),  # type: ignore
                prompt_template_name=json_dom.get('prompt_template_name'),  # type: ignore
                content=FilesystemTemplateResolver.__normalize_roles(json_dom['content']),
                metadata=PromptTemplateMetadata(
                    provider=FilesystemTemplateResolver.__flavor_to_provider(flavor_name),
                    flavor=flavor_name,
                    model=model,
                    params=metadata.get('params'),
                    provider_info=metadata.get('provider_info')
                )
            )
        else:
            metadata = json_dom['metadata']

            flavor_name = metadata.get('flavor_name')
            params = metadata.get('params')
            model = params.pop('model') if 'model' in params else None

            return PromptTemplate(
                format_version=2,
                prompt_template_id=json_dom.get('prompt_template_id'),  # type: ignore
                prompt_template_version_id=json_dom.get('prompt_template_version_id'),  # type: ignore
                prompt_template_name=json_dom.get('name'),  # type: ignore
                content=FilesystemTemplateResolver.__normalize_roles(json.loads(str(json_dom['content']))),
                metadata=PromptTemplateMetadata(
                    provider=FilesystemTemplateResolver.__flavor_to_provider(flavor_name),
                    flavor=flavor_name,
                    model=model,
                    params=params,
                    provider_info=None
                )
            )

    @staticmethod
    def __normalize_roles(messages: List[Dict[str, str]]) -> List[Dict[str, str]]:
        normalized = []
        for message in messages:
            role = FilesystemTemplateResolver.__role_translations.get(message['role']) or message['role']
            normalized.append({'role': role, 'content': message['content']})
        return normalized

    @staticmethod
    def __validate_freeplay_directory(freeplay_directory: Path) -> None:
        if not freeplay_directory.is_dir():
            raise FreeplayConfigurationError(
                "Path for prompt templates is not a valid directory (%s)" % freeplay_directory
            )

        prompts_directory = freeplay_directory / "freeplay" / "prompts"
        if not prompts_directory.is_dir():
            raise FreeplayConfigurationError(
                "Invalid path for prompt templates (%s). "
                "Did not find a freeplay/prompts directory underneath." % freeplay_directory
            )

    def __validate_prompt_directory(self, project_id: str, environment: str) -> None:
        maybe_prompt_dir = self.prompts_directory / project_id / environment
        if not maybe_prompt_dir.is_dir():
            raise FreeplayConfigurationError(
                "Could not find prompt template directory for project ID %s and environment %s." %
                (project_id, environment)
            )

    @staticmethod
    def __flavor_to_provider(flavor: str) -> str:
        flavor_provider = {
            'azure_openai_chat': 'azure',
            'anthropic_chat': 'anthropic',
            'openai_chat': 'openai',
        }
        provider = flavor_provider.get(flavor)
        if not provider:
            raise MissingFlavorError(flavor)
        return provider


class APITemplateResolver(TemplateResolver):

    def __init__(self, call_support: CallSupport):
        self.call_support = call_support

    def get_prompts(self, project_id: str, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts(
            project_id=project_id,
            environment=environment
        )

    def get_prompt(self, project_id: str, template_name: str, environment: str) -> PromptTemplate:
        return self.call_support.get_prompt(
            project_id=project_id,
            template_name=template_name,
            environment=environment
        )


class Prompts:
    def __init__(self, call_support: CallSupport, template_resolver: TemplateResolver) -> None:
        self.call_support = call_support
        self.template_resolver = template_resolver

    def get_all(self, project_id: str, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts(project_id=project_id, environment=environment)

    def get(self, project_id: str, template_name: str, environment: str) -> TemplatePrompt:
        prompt = self.template_resolver.get_prompt(project_id, template_name, environment)

        params = prompt.metadata.params
        model = prompt.metadata.model

        if not model:
            raise FreeplayConfigurationError(
                "Model must be configured in the Freeplay UI. Unable to fulfill request.")

        if not prompt.metadata.flavor:
            raise FreeplayConfigurationError(
                "Flavor must be configured in the Freeplay UI. Unable to fulfill request.")

        if not prompt.metadata.provider:
            raise FreeplayConfigurationError(
                "Provider must be configured in the Freeplay UI. Unable to fulfill request.")

        prompt_info = PromptInfo(
            prompt_template_id=prompt.prompt_template_id,
            prompt_template_version_id=prompt.prompt_template_version_id,
            template_name=prompt.prompt_template_name,
            environment=environment,
            model_parameters=cast(LLMParameters, params) or LLMParameters({}),
            provider=prompt.metadata.provider,
            model=model,
            flavor_name=prompt.metadata.flavor,
            provider_info=prompt.metadata.provider_info
        )

        return TemplatePrompt(prompt_info, prompt.content)

    def get_formatted(
            self,
            project_id: str,
            template_name: str,
            environment: str,
            variables: InputVariables,
            flavor_name: Optional[str] = None
    ) -> FormattedPrompt:
        bound_prompt = self.get(
            project_id=project_id,
            template_name=template_name,
            environment=environment
        ).bind(variables=variables)

        return bound_prompt.format(flavor_name)



import json
import os
import sys
from pathlib import Path
from stat import S_IREAD, S_IRGRP, S_IROTH, S_IWUSR

import click

from freeplay.errors import FreeplayClientError, FreeplayServerError
from freeplay import Freeplay
from freeplay.support import PromptTemplates, PromptTemplate, PromptTemplateEncoder


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option("--project-id", required=True, help="The Freeplay project ID.")
@click.option("--environment", required=True, help="The environment from which the prompts will be pulled.")
@click.option("--output-dir", required=True, help="The directory where the prompts will be saved.")
def download(project_id: str, environment: str, output_dir: str) -> None:
    if "FREEPLAY_API_KEY" not in os.environ:
        print("FREEPLAY_API_KEY is not set. It is required to run the freeplay command.", file=sys.stderr)
        exit(4)

    if "FREEPLAY_SUBDOMAIN" not in os.environ:
        print("FREEPLAY_SUBDOMAIN is not set. It is required to run the freeplay command.", file=sys.stderr)
        exit(4)

    FREEPLAY_API_KEY = os.environ["FREEPLAY_API_KEY"]
    freeplay_api_url = f'https://{os.environ["FREEPLAY_SUBDOMAIN"]}.freeplay.ai/api'

    if "FREEPLAY_API_URL" in os.environ:
        freeplay_api_url = f'{os.environ["FREEPLAY_API_URL"]}/api'
        click.echo("Using URL override for Freeplay specified in the FREEPLAY_API_URL environment variable")

    click.echo(
        f"Downloading prompts for project {project_id}, environment {environment}, "
        f"to directory {output_dir} from {freeplay_api_url}")

    fp_client = Freeplay(
        freeplay_api_key=FREEPLAY_API_KEY,
        api_base=freeplay_api_url
    )

    try:
        prompts: PromptTemplates = fp_client.prompts.get_all(project_id, environment=environment)
        click.echo(f"Found {len(prompts.prompt_templates)} prompt templates")

        for prompt in prompts.prompt_templates:
            __write_single_file(environment, output_dir, project_id, prompt)
    except FreeplayClientError as e:
        print(f"Error downloading templates: {e}.\nIs your project ID correct?", file=sys.stderr)
        exit(1)
    except FreeplayServerError as e:
        print(f"Error on Freeplay's servers downloading templates: {e}.\nTry again after a short wait.",
              file=sys.stderr)
        exit(2)
    except Exception as e:
        print(f"Error downloading templates: {e}", file=sys.stderr)
        exit(3)


def __write_single_file(
        environment: str,
        output_dir: str,
        project_id: str,
        prompt: PromptTemplate
) -> None:
    directory = __root_dir(environment, output_dir, project_id)
    basename = f'{prompt.prompt_template_name}'
    prompt_path = directory / f'{basename}.json'
    click.echo("Writing prompt file: %s" % prompt_path)

    # Make sure it's owner writable if it already exists
    if prompt_path.is_file():
        os.chmod(prompt_path, S_IWUSR | S_IREAD)

    with prompt_path.open(mode='w') as f:
        f.write(json.dumps(prompt, sort_keys=True, indent=4, cls=PromptTemplateEncoder))
        f.write('\n')

    # Make the file read-only to discourage local changes
    os.chmod(prompt_path, S_IREAD | S_IRGRP | S_IROTH)


def __root_dir(environment: str, output_dir: str, project_id: str) -> Path:
    directory = Path(output_dir) / "freeplay" / "prompts" / project_id / environment
    os.makedirs(directory, exist_ok=True)
    return directory
